#===============================================================================
#
# Copyright (c) 2014 Qualcomm Technologies, Inc. All Rights Reserved.
# Qualcomm Technologies Proprietary and Confidential.
#
#===============================================================================

import struct
import defines

from sectools.features.fuse_common.parsegen.secdat.secdatbase import SecdatParseGen, fuseprov_secdat_footer_type
from sectools.common.utils.c_logging import logger

class SecdatParseGenRevTwo(SecdatParseGen):

    def __init__(self, header_info='', header_version=None,
                            fuse_list_header_version=None):
        SecdatParseGen.__init__(self, header_info, header_version,
                                fuse_list_header_version)


    #--------------------------------------------------------------------------
    # Mandatory Properties overloaded from the base class
    #--------------------------------------------------------------------------
    @classmethod
    def is_plugin(cls):
        return True

    @classmethod
    def file_version(cls):
        return defines.SECDAT_REVISION_TWO

    def modifyhdr(self, cmn_hdr_data, size):

        # Modify header data as part of generating individual data model for each segment
        data = struct.unpack(defines.HDR_FMT, cmn_hdr_data)
        cmn_hdr_data = struct.pack(defines.HDR_FMT, data[0], data[1], data[2], size, data[4], data[5], data[6], data[7], data[8], data[9], data[10], data[11])
        return cmn_hdr_data

    def modifyseginfo(self, seg_info_data):

        # Modify segment information data in header as part of generating individual data model for each segment
        data = struct.unpack(defines.SEGINFO_FMT, seg_info_data)
        seg_info_data = struct.pack(defines.SEGINFO_FMT, (defines.SEG_START_ADDR), data[1])
        return seg_info_data

    def validate_header(self, header, file_size):

        if not (header.magic1 == defines.FUSEPROV_SECDAT_MAGIC1 and
                header.magic2 == defines.FUSEPROV_SECDAT_MAGIC2 and
                header.revision == defines.SECDAT_REVISION_TWO and
                (header.size + header.PACK_SIZE == file_size)):
            raise RuntimeError('Header Mismatch')
        pass

    def parse_secdat_hdr_data(self, file_data):

        # Extract the header data from the sec.dat file
        file_data_len = len(file_data)

        header_data = file_data[:_fuseprov_secdat_hdr_type.PACK_SIZE]
        file_data = file_data[_fuseprov_secdat_hdr_type.PACK_SIZE:]
        self._header = _fuseprov_secdat_hdr_type(header_data)
        self.validate_header(self._header, file_data_len)
        self.header_info = self._header.info
        self.header_version = self._header.revision
        return header_data, file_data

    def create_header(self, data_size, info, feature):

        #--------------------------------------------------------------------------------------------------------
        # Frame the header for sec.dat revision-2
        # Header creation which includes single segment either fuseblower or keyprovision
        #--------------------------------------------------------------------------------------------------------
        header = _fuseprov_secdat_hdr_type()
        header.magic1 = defines.FUSEPROV_SECDAT_MAGIC1
        header.magic2 = defines.FUSEPROV_SECDAT_MAGIC2
        header.revision = defines.SECDAT_REVISION_TWO
        header.size = data_size
        header.info = info.ljust(defines.FUSEPROV_INFO_MAX_SIZE, '\0')
        header.segnum = defines.ONE_SEGMENT
        header.segoff = defines.ONE_SEG_OFFSET

        if feature is defines.FUSEBLOWER:
            header.segtype = defines.EFUSE_SEGMENT
        elif feature is defines.KEYPROVISION:
            header.segtype = defines.KEYPROVISION_SEGMENT
        else:
            raise RuntimeError('Segment type can not be other than efuse & keyprovision')

        header.segattr = defines.DEFAULT_SEG_ATTR
        return header
        #--------------------------------------------------------------------------------------------------------

    def create_newhdr(self, data_size, info, hdr_data, feature):

        #--------------------------------------------------------------------------------------------------------
        # Frame the header for sec.dat revision-2
        # Header creation which includes two segments fuseblower & keyprovision
        #--------------------------------------------------------------------------------------------------------
        header = _fuseprov_newsecdat_hdr_type()
        header.magic1 = defines.FUSEPROV_SECDAT_MAGIC1
        header.magic2 = defines.FUSEPROV_SECDAT_MAGIC2
        header.revision = defines.SECDAT_REVISION_TWO
        header.size = data_size
        header.info = info.ljust(defines.FUSEPROV_INFO_MAX_SIZE, '\0')
        header.segnum = defines.TWO_SEGMENTS
        header.seg1off = hdr_data.segoff + 8
        header.seg1type = hdr_data.segtype
        header.seg1attr = hdr_data.segattr
        header.seg2off = hdr_data.size + defines.TWO_SEGS_HEADER_SIZE - defines.FOOTER_SIZE

        if feature is defines.FUSEBLOWER:
            header.seg2type = defines.EFUSE_SEGMENT
        elif feature is defines.KEYPROVISION:
            header.seg2type = defines.KEYPROVISION_SEGMENT
        else:
            raise RuntimeError('Segment type can not be other than efuse & keyprovision')

        header.seg2attr = defines.DEFAULT_SEG_ATTR
        return header
        #--------------------------------------------------------------------------------------------------------

    def update(self, data_model):

        # Packs the fuse header and fuse data from into single data from the given data model
        self._fuse_list = []
        for module in data_model.module:
            for fuse_region in module.fuse_region:
                for fuse in fuse_region.fuse:
                    if fuse.ignore:
                        logger.info('Ignoring fuse: ' + str(fuse))
                        continue
                    fuse_entry = self.create_fuse_entry(self.get_region_enum_from_id(fuse_region.id),
                                                        fuse.address,
                                                        self.get_operation_enum_from_id(fuse.operation),
                                                        fuse.value)
                    self.add_fuse(fuse_entry)


        self._fuse_header = self.create_fuse_header(len(self._fuse_list))
        data = self._fuse_header.get_packed_data()
        for fuse in self._fuse_list:
            data += fuse.get_packed_data()

        return data

    def secdat_input_parser(self, input_secdat, feature):

        self.fb_dm = None
        self.kp_dm = None

        if input_secdat:
            self.readsecdatfile(input_secdat)

            if self.secdatrevision == defines.SECDAT_REVISION_ONE:
                return self.fb_dm, self.kp_dm

            else:
                if self.fb_seg is not None:
                    # Set fuseblower data_model, header, footer, fuse_hdr, fuse_data from fb_seg
                    self.parse(self.fb_seg)
                    self.update_data_model()
                    self.fb_dm = self.data_model
                    self.fb_hdr = self._header
                    self.fb_ftr = self._footer
                    self.fb_fhdr = self._fuse_header
                    self.fb_fuse = self._fuse_list
                if self.kp_seg is not None:
                    # Set keyprovisioin data_model, header, footer, fuse_hdr, fuse_data from kp_seg
                    self.parse(self.kp_seg)
                    self.update_data_model()
                    self.kp_dm = self.data_model
                    self.kp_hdr = self._header
                    self.kp_ftr = self._footer
                    self.kp_fhdr = self._fuse_header
                    self.kp_fuse = self._fuse_list

                return self.fb_dm, self.kp_dm

        else:
            self.secdatrevision = defines.SECDAT_REVISION_TWO
            return self.fb_dm, self.kp_dm

    def readfirstseg_secdat(self, fd):

        # Extract segment header data, fuse header & fuse data, footer from the given file pointer
        fd.seek(0)
        self.cmn_hdr_data = fd.read(48)
        seg_info_data = fd.read(8)
        fd.seek(0)
        fd.seek(self.secdat_hdr.seg1off)
        seg_fuse_data = fd.read(self.secdat_hdr.seg2off - self.secdat_hdr.seg1off)
        self.cmn_hdr_data = self.modifyhdr(self.cmn_hdr_data, len(seg_fuse_data)+ defines.FOOTER_SIZE )
        seg_info_data = self.modifyseginfo(seg_info_data)
        self.seg_data = self.cmn_hdr_data + seg_info_data + seg_fuse_data
        seg_ftr_obj = self.create_footer(self.seg_data)
        self.seg_ftr_data = seg_ftr_obj.hash

    def readsecondseg_secdat(self, fd, datasize):

        # Extract segment header data, fuse header & fuse data, footer from the given file pointer
        fd.seek(0)
        fd.seek(56)
        seg_info_data = fd.read(8)
        fd.seek(0)
        fd.seek(self.secdat_hdr.seg2off)
        seg_fuse_data = fd.read(datasize - self.secdat_hdr.seg2off - defines.FOOTER_SIZE)
        self.cmn_hdr_data = self.modifyhdr(self.cmn_hdr_data, len(seg_fuse_data)+ defines.FOOTER_SIZE )
        seg_info_data = self.modifyseginfo(seg_info_data)
        self.seg_data = self.cmn_hdr_data + seg_info_data + seg_fuse_data
        seg_ftr_obj = self.create_footer(self.seg_data)
        self.seg_ftr_data = seg_ftr_obj.hash

    def readsecdatfile(self, file_path):

        #--------------------------------------------------------------------------------------------------------
        # Read the given sec.dat file
        # If two segments are available in sec.dat, then extract the two segments data separately.
        # Create individual segment header + fuse header + fuse data + segment footer for fuseblower & keyprovision
        #--------------------------------------------------------------------------------------------------------
        fd = open(file_path, 'rb')
        file_data = fd.read()

        datasize = len(file_data)
        secdat_hdr_data = file_data[:_fuseprov_secdat_hdr_type.PACK_SIZE]
        secdat_hdr = _fuseprov_secdat_hdr_type(secdat_hdr_data)

        self.fb_seg = None
        self.kp_seg = None

        if secdat_hdr.revision == defines.SECDAT_REVISION_TWO:
            self.secdatrevision = defines.SECDAT_REVISION_TWO
            if secdat_hdr.segnum == defines.TWO_SEGMENTS:
                # sec.dat header revision is 2 and two segments are available
                secdat_hdr_data = file_data[:_fuseprov_newsecdat_hdr_type.PACK_SIZE]
                self.secdat_hdr = _fuseprov_newsecdat_hdr_type(secdat_hdr_data)

                if self.secdat_hdr.seg1type == defines.EFUSE_SEGMENT:
                    if self.secdat_hdr.seg2type == defines.KEYPROVISION_SEGMENT:
                        # Read fuseblower segment as segment 1 (fb_seg) & keyprovision segment as segment 2 (kp_seg)
                        self.readfirstseg_secdat(fd)
                        self.fb_seg = self.seg_data + self.seg_ftr_data
                        self.readsecondseg_secdat(fd, datasize)
                        self.kp_seg = self.seg_data + self.seg_ftr_data

                elif self.secdat_hdr.seg1type == defines.KEYPROVISION_SEGMENT:
                    if self.secdat_hdr.seg2type == defines.EFUSE_SEGMENT:
                        # Read keyprovision segment as segment 1 (kp_seg) & fuseblower segment as segment 2 (fb_seg)
                        self.readfirstseg_secdat(fd)
                        self.kp_seg = self.seg_data + self.seg_ftr_data
                        self.readsecondseg_secdat(fd, datasize)
                        self.fb_seg = self.seg_data + self.seg_ftr_data

                else:
                    fd.close()
                    raise RuntimeError('Segment type can not be other than efuse & keyprovision')

                fd.close()


            elif secdat_hdr.segnum == defines.ONE_SEGMENT:
                # sec.dat header revision is 2 and only segment is available
                if secdat_hdr.segtype == defines.EFUSE_SEGMENT:
                    # Assign fuseblower segment to fb_seg
                    self.fb_seg = file_data
                elif secdat_hdr.segtype == defines.KEYPROVISION_SEGMENT:
                    # Assign keyprovision segment to kp_seg
                    self.kp_seg = file_data
                else:
                    fd.close()
                    raise RuntimeError('Segment type can not be other than efuse & keyprovision')
                fd.close()

        elif secdat_hdr.revision == defines.SECDAT_REVISION_ONE:
            self.secdatrevision = defines.SECDAT_REVISION_ONE
            fd.close()

        else:
            fd.close()
            raise RuntimeError('Sec.dat revision can not be other than one & two')


    def write(self, data_model, input_secdat, secdat, common_secdat, exist_secdat, feature):

        if input_secdat is None:
            self.update_from_data_model(data_model, feature)
            self.oneseghdr = self._header
            self.onesegftr = self._footer
            self.twoseghdr = None
            self.twosegftr = None
            self.generate(secdat)
            return

        else:
            # fuseblower feature
            if feature == 'fuseblower':
                if self.fb_dm is not None:
                    if self.kp_dm is not None:
                        old_segdata = self.update(self.kp_dm)
                        hdr_data = self.kp_hdr
                        self.oneseghdr = self.kp_hdr
                        self.onesegftr = self.kp_ftr
                    else:
                        raise RuntimeError('Passing Fuseblower_sec.dat as input to Fuseblower for merging' '\n'
                                   'Please remove sec.dat as input and try again')

                else:
                    if self.kp_dm is not None:
                        old_segdata = self.update(self.kp_dm)
                        hdr_data = self.kp_hdr
                        self.oneseghdr = self.kp_hdr
                        self.onesegftr = self.kp_ftr
                    else:
                        if self.secdatrevision == defines.SECDAT_REVISION_ONE:
                            if exist_secdat is None:
                                raise RuntimeError('Passing fuseblower sec.dat revision-1 as input to Fuseblower_Sec.dat_Revision-2' '\n'
                                                   'Please remove sec.dat as input and try again')
                            else:
                                raise RuntimeError('Both the data models can not be None' '\n')
                        else:
                            raise RuntimeError('Both the data models can not be None' '\n')

            # keyprovision feature
            else:
                if self.kp_dm is not None:
                    if self.fb_dm is not None:
                        old_segdata = self.update(self.fb_dm)
                        hdr_data = self.fb_hdr
                        self.oneseghdr = self.fb_hdr
                        self.onesegftr = self.fb_ftr
                    else:
                        raise RuntimeError('Passing Keyprovision_sec.dat as input to Keyprovision for merging' '\n'
                                   'Please remove sec.dat as input and try again')

                else:
                    if self.fb_dm is not None:
                        old_segdata = self.update(self.fb_dm)
                        hdr_data = self.fb_hdr
                        self.oneseghdr = self.fb_hdr
                        self.onesegftr = self.fb_ftr
                    else:
                        if self.secdatrevision == defines.SECDAT_REVISION_ONE:
                            if exist_secdat is None:
                                raise RuntimeError('Passing fuseblower sec.dat revision-1 as input to Keyprovision_Sec.dat_Revision-2' '\n'
                                                   'Please remove sec.dat as input and try again')
                            else:
                                self.update_from_data_model(data_model, feature)
                                self.oneseghdr = self._header
                                self.onesegftr = self._footer
                                self.twoseghdr = None
                                self.twosegftr = None
                                self.generate(secdat)
                                return
                        else:
                            raise RuntimeError('Both the data models can not be None' '\n')


            new_segdata = self.update(data_model)


            self._header = self.create_newhdr(len(old_segdata) + len(new_segdata) + fuseprov_secdat_footer_type.PACK_SIZE, self.header_info, hdr_data, feature)
            data = self._header.get_packed_data() + old_segdata + new_segdata
            self._footer = self.create_footer(data)

            file_data = self._header.get_packed_data() + old_segdata + new_segdata + self._footer.get_packed_data()

            self.twoseghdr = self._header
            self.twosegftr = self._footer
            fd = open(common_secdat, 'wb')
            fd.write(file_data)
            fd.close()

            if exist_secdat is None:
                return

            else:
                self.update_from_data_model(data_model, feature)
                self.oneseghdr = self._header
                self.onesegftr = self._footer
                self.generate(secdat)


    def secdat_repr(self, file_path, segs, feature):

        sec_data = ''

        if segs == defines.ONE_SEGMENT:
            self._header = self.oneseghdr
            self._footer = self.onesegftr

        elif segs == defines.TWO_SEGMENTS:
            self._header = self.twoseghdr
            self._footer = self.twosegftr
            if feature == defines.FUSEBLOWER:
                if self.kp_dm:
                    self.fhdr = self.kp_fhdr
                    self.fuse = self.kp_fuse
            else:
                if self.fb_dm:
                    self.fhdr = self.fb_fhdr
                    self.fuse = self.fb_fuse

            sec_data += ('Fuse Header: ' + '\n'
                             '' + logger.increase_indent(repr(self.fhdr)) + '\n')

            for fuse in self.fuse:
                sec_data += ('' + 'Fuse: ' + '\n'
                                 '' + logger.increase_indent(repr(fuse)) + '\n')

        else:
            raise RuntimeError('segments supported are 1 and 2, got value as : ' + str(segs))

        retval = ''
        retval += ('Secdat properties: ' + '\n' + '\n'
                   '' + 'File path: ' + file_path + '\n' + '\n'
                   '' + 'Header: ' + '\n'
                   '' + logger.increase_indent(repr(self._header)) + '\n'
                   '' + sec_data  + '\n'
                   '' + 'Fuse Header: ' + '\n'
                   '' + logger.increase_indent(repr(self._fuse_header)) + '\n')

        for fuse in self._fuse_list:
            retval += ('' + 'Fuse: ' + '\n'
                       '' + logger.increase_indent(repr(fuse)) + '\n')

        retval += ('Footer: ' + '\n'
                   '' + logger.increase_indent(repr(self._footer)) + '\n')


        return retval.strip()


class _fuseprov_secdat_hdr_type(object):

    # Structure object to align and package the header
    PACK_FMT = struct.Struct('IIII' +
                             'B' * defines.FUSEPROV_INFO_MAX_SIZE +
                             'I' +
                             'I' * 3 +
                             'I' +
                             'H' * 2)

    # Size of the header
    PACK_SIZE = ((7 * 4) +
                (1 * defines.FUSEPROV_INFO_MAX_SIZE) +
                (4 * 2)) + (2 * 2)


    def __init__(self, data=None):
        if data is None:
            data = ('{0:c}'.format(0))*self.PACK_SIZE
            #print data
        unpacked_data = (self.PACK_FMT).unpack(data)
        #print unpacked_data
        self.magic1 = unpacked_data[0]
        self.magic2 = unpacked_data[1]
        self.revision = unpacked_data[2]
        self.size = unpacked_data[3]
        self.info = ''
        self.segnum = unpacked_data[20]
        self.reserved = ''
        self.segoff = unpacked_data[24]
        self.segtype = unpacked_data[25]
        self.segattr = unpacked_data[26]

        # Pull info
        i = 4
        for i in range(i, i + defines.FUSEPROV_INFO_MAX_SIZE):
            self.info += '{0:c}'.format(unpacked_data[i])
        # Pull reserved
        i = 21
        for i in range(i, i + 3):
            self.reserved += '{0:c}'.format(unpacked_data[i])
        pass

    def __repr__(self):
        retval = ''
        for attr in sorted(self.__dict__.keys()):
            value = self.__dict__[attr]
            retval += (str(attr) + ': ' + (value if isinstance(value, str) else
                       str(hex(value))) + '\n')
        return (self.__class__.__name__ + ' Properties: ' + '\n' +
                logger.increase_indent(retval))

    def get_packed_data(self):
        values = ([self.magic1,
                   self.magic2,
                   self.revision,
                   self.size] +
                  [ord(c) for c in self.info] +
                  [self.segnum]+
                  [ord(c) for c in self.reserved]+
                  [self.segoff]+
                  [self.segtype,
                   self.segattr]
                 )
        return (self.PACK_FMT).pack(*values)


class _fuseprov_newsecdat_hdr_type(object):

    # Structure object to align and package the header
    PACK_FMT = struct.Struct('IIII' +
                             'B' * 16 +
                             'I' +
                             'I' * 3 +
                             'I' +
                             'H' * 2 +
                             'I'+
                             'H' * 2)

    # Size of the header
    PACK_SIZE = ((7 * 4) +
                (1 * 16) +
                (4 * 3)) + (4 * 2)


    def __init__(self, data=None):
        if data is None:
            data = ('{0:c}'.format(0))*self.PACK_SIZE
            #print data
        unpacked_data = (self.PACK_FMT).unpack(data)
        #print unpacked_data
        self.magic1 = unpacked_data[0]
        self.magic2 = unpacked_data[1]
        self.revision = unpacked_data[2]
        self.size = unpacked_data[3]
        self.info = ''
        self.segnum = unpacked_data[20]
        self.reserved = ''
        self.seg1off = unpacked_data[24]
        self.seg1type = unpacked_data[25]
        self.seg1attr = unpacked_data[26]
        self.seg2off = unpacked_data[27]
        self.seg2type = unpacked_data[28]
        self.seg2attr = unpacked_data[29]


        # Pull info
        i = 4
        for i in range(i, i + 16):
            self.info += '{0:c}'.format(unpacked_data[i])
        # Pull reserved
        i = 21
        for i in range(i, i + 3):
            self.reserved += '{0:c}'.format(unpacked_data[i])
        pass

    def __repr__(self):
        retval = ''
        for attr in sorted(self.__dict__.keys()):
            value = self.__dict__[attr]
            retval += (str(attr) + ': ' + (value if isinstance(value, str) else
                       str(hex(value))) + '\n')
        return (self.__class__.__name__ + ' Properties: ' + '\n' +
                logger.increase_indent(retval))

    def get_packed_data(self):
        values = ([self.magic1,
                   self.magic2,
                   self.revision,
                   self.size] +
                  [ord(c) for c in self.info] +
                  [self.segnum]+
                  [ord(c) for c in self.reserved]+
                  [self.seg1off]+
                  [self.seg1type,
                   self.seg1attr]+
                  [self.seg2off]+
                  [self.seg2type,
                   self.seg2attr]
                 )
        return (self.PACK_FMT).pack(*values)
